import os
import numpy as np
import pandas as pd

from datetime import date
from datetime import datetime
from datetime import timedelta
from typing import Union
from sqlalchemy.types import DATE
from sqlalchemy.types import DECIMAL
from sqlalchemy.types import INT
from sqlalchemy.types import NVARCHAR
from sqlalchemy.types import VARCHAR
from sqlalchemy.types import Boolean
from acechmscoring import sql, utils

MODEL_VERSION = os.environ.get("MODEL_VERSION")
SALES_ORG_NAME = os.environ.get("SALES_ORG_NAME")

def get_p_history_update(
    table_name: str,
    max_sales: datetime,
    bw: bool = False,
    model_version: str = MODEL_VERSION,
    sales: pd.DataFrame = None,
    verbose: bool = True,
) -> pd.DataFrame:
    """Get P history SQL tables up to a given max_sales date.

    Args:
        table_name: Probably CHM_P_alive or CHM_P_purchase
        max_sales: The max date of sales data. Usually obtained from
            sql.get_max_sales_date().
        bw: Query BW instead of Sherlock.
        sales: This is really to only be used when calling this method twice in
            a row with the same parameters, to get P_alive and P_purchase.
            Return the downloaded sales data in the first call, and pass it
            through in the second call to avoid downloading the same data
            twice.

    """
    max_p = sql.get_max("Date", table_name)
    if max_p is None:
        if verbose:
            print("ERROR: max_p date is None.")
        return (None, None)

    if table_name.startswith("CHM_P_alive"):
        model = sql.get_model_alive(model_version)
        p_col = "P_alive"
    elif table_name.startswith("CHM_P_purchase"):
        model = sql.get_model_purchase(model_version)
        p_col = "P_purchase"
    else:
        if verbose:
            print("ERROR: Invalid table_name.")
        return (None, None)

    if max_p < max_sales:
        min_date = max_p + timedelta(days=1)
        max_date = max_sales

        if sales is None:
            sales = sql.read(
                f"""
                SELECT
                    CALDAY,
                    ZCUSTOMER,
                    BILL_TYPE,
                    PROD_SALES_LC AS NetSale,
                    PROD_SALES_QTY AS NetUnits
                FROM view_CHM_CopaSales AS A
                LEFT JOIN SF_BrandAccount AS C
                  ON A.ZCUSTOMER = C.ERP_Account_Number__c
                  AND A.SalesOrg_Name = C.SalesOrg_Name
                WHERE A.SALESORG = '{sql.get_salesorg()}'
                  {sql.get_territory_filter("C.", "SalesGroup__c")}
                  AND A.CALDAY >= '{min_date.strftime("%Y%m%d")}'
                  AND A.CALDAY <= '{max_date.strftime("%Y%m%d")}'
                """,
                show_query=False,
            )
            sales["CALDAY"] = pd.to_datetime(sales["CALDAY"])
            sales = sales.rename(columns={"CALDAY": "Date"})

        if table_name.startswith("CHM_P_purchase"):
            sales = utils.get_purchases(sales)
        last_row = sql.read(
            f"""
            SELECT ZCUSTOMER, T, frequency, recency
            FROM {table_name}
            WHERE [Date] = '{max_p.strftime("%Y-%m-%d")}'
              AND SalesOrg_Name = '{SALES_ORG_NAME}'
            """,
            show_query=False,
        ).set_index("ZCUSTOMER")
        return (
            build_p_history_dataframe(
                sales=sales,
                model=model,
                model_version=model_version,
                p_col=p_col,
                id_col="ZCUSTOMER",
                datetime_col="Date",
                start_date=min_date,
                end_date=max_date,
                append_row=last_row,
                ignore_first_purchase=False,
            ),
            sales,
        )

    if verbose:
        print(f"{table_name} is already updated until {max_sales}.")
    return (None, None)


def scoring(max_sales_date: date = None):
    t_start = datetime.now()
    DTYPE = {
        "ZCUSTOMER": VARCHAR(50),
        "SalesOrg_Name": NVARCHAR(240),
        "Date": DATE(),
        "transactions": Boolean(),
        "T": INT(),
        "frequency": INT(),
        "recency": INT(),
        "P_purchase": DECIMAL(18, 17),
        "P_alive": DECIMAL(18, 17),
        "ModelVersion": VARCHAR(50),
    }

    suffix = ""
    if max_sales_date is None:
        max_sales_date = sql.get_max_sales_date()
    date_range = [max_sales_date]

    for max_sales in date_range:
        p_history_alive, sales = get_p_history_update(f"CHM_P_alive{suffix}", max_sales)
        if p_history_alive is None or sales is None:
            continue

        p_history_purchase, _ = get_p_history_update(
            f"CHM_P_purchase{suffix}", max_sales, sales=sales
        )

        sql.append(f"CHM_P_alive{suffix}", p_history_alive, dtype=DTYPE)
        sql.append(f"CHM_P_purchase{suffix}", p_history_purchase, dtype=DTYPE)
    update_p_min(suffix=suffix)
    print(f"scoring.py elapsed: {datetime.now() - t_start}")

def update_p_min(suffix: str = "", verbose: bool = True):
    max_p_alive = sql.get_max("[Date]", f"CHM_P_alive{suffix}")
    max_p_purchase = sql.get_max("[Date]", f"CHM_P_purchase{suffix}")
    max_p_min = sql.get_max("[Date]", f"CHM_P_min{suffix}")

    if max_p_min < min([max_p_alive, max_p_purchase]):
        min_date = (max_p_min + timedelta(days=1)).strftime("%Y-%m-%d")

        sql.execute(
            f"""
            INSERT INTO CHM_P_min{suffix}
            SELECT
                A.ZCUSTOMER,
                A.SalesOrg_Name,
                A.[Date],
                B.T - B.recency AS Recency,
                CASE
                    WHEN B.P_purchase < A.P_alive
                    THEN B.P_purchase
                    ELSE A.P_alive
                END AS P_min
            FROM CHM_P_alive AS A
            FULL OUTER JOIN CHM_P_purchase AS B
                ON A.ZCUSTOMER = B.ZCUSTOMER
                AND A.SalesOrg_Name = B.SalesOrg_Name
                AND A.[Date] = B.[Date]
            WHERE A.ZCUSTOMER IS NOT NULL
                AND B.ZCUSTOMER IS NOT NULL
                AND A.SalesOrg_Name = '{SALES_ORG_NAME}'
                AND B.SalesOrg_Name = '{SALES_ORG_NAME}'
                AND A.[Date] >= '{min_date}'
                AND B.[Date] >= '{min_date}'
            """
        )
        if verbose:
            print(f"Done updating CHM_P_min{suffix}")

    else:
        if verbose:
            print("Nothing to update.")

    return

def build_p_history_dataframe(
    sales: pd.DataFrame,
    model=None,
    tail=None,
    p_col: str = "P_alive",
    datetime_col: str = "Date",
    id_col: str = "SoldTo",
    model_version: str = "unknown",
    start_date: Union[int, datetime] = None,
    end_date: Union[int, datetime] = None,
    append_row: pd.Series = None,
    ignore_first_purchase: bool = True,
    filename: str = None,
    verbose: bool = True,
) -> pd.DataFrame:
    """Calculate the history of p values for all accounts.

    This is effectively a refactored version of
    lifetimes.utils.calculate_alive_path() that operates on all accounts'
    transactions at once.

    Args:
        sales (pandas.DataFrame): Transactional dataframe.
        model: Lifetimes model to add as column.
        tail (int or datetime): Filter the output to only include this number
            of days back from the max date in the dataframe. Alternatively,
            pass a datetime to filter greater than or equal to that date. This
            is done before calculating probabilities.
        p_col (str): Name for the probability column created from model.
            Default "P_alive".
        datetime_col (str): Default "Date".
        id_col (str): Default "SoldTo". Will soon be replaced with "ZCUSTOMER".
        start_date (int or datetime): First date of the data, to where you would
            like to draw the p line.
        end_date (int or datetime): Last date of the data, to where you would
            like to draw the p line.
        append_row (pandas.Series): Data to use as the first row of the summary
            dataframe, before adding the trnasactions from sales. Should have
            index <id_col>, and columns T, frequency, and recency. This is
            intended to be used when appending the result to an existing table.
            This argument should be the last row from the target table from
            which to add the new values for T, frequency, and recency.
        filename (str, optional): Optional path to save the resulting dataframe
            as parquet.

    Returns:
        pandas.DataFrame: A dataframe with index "SoldTo" and "Date" and
        columns for T, frequency, and recency (age, transaction-days, and
        tenure). Additional columns for model output are included if models
        are specified. The "IsPuchase" column is also kept from the original
        sales dataframe.

    """
    if verbose:
        timer = datetime.now()
        print(f"Calculating {p_col} history...")
        print(f"  Started: {timer}")

    p_history = sales.reset_index().set_index(datetime_col)[[id_col]]
    p_history["transactions"] = 1
    end_date = p_history.index.max() if end_date is None else end_date
    start_date = p_history.index.min() if start_date is None else start_date
    end_date = pd.to_datetime(end_date)
    start_date = pd.to_datetime(start_date)

    def get_date(the_date, exclude_ids: list = None, include_ids: list = None):
        """Get a row of a date that can be appended to a dataframe.
        Assumes that the index of p_history is datetime.

        Args:
            exclude_ids: A list of ids to exclude from the result. Needed to
                avoid generating probabilities for days prior to the first
                transaction of new accounts.
            include_ids: A list of ids to append to the result. If the ids
                already exist, they will end up being duplicated. Needed to
                generate probabilites for accounts that didn't have any
                transactions in the update period. This is applied *after*
                exclude_ids, so an id appearing in both list will end up being
                included.
        """
        if the_date in p_history.index:
            # get list of ids without a row for the_date
            has_transaction_at_date = p_history.xs(the_date)[id_col]
            no_transaction_at_date = ~p_history[id_col].isin(
                list(has_transaction_at_date)
            )
            append_ids = p_history[no_transaction_at_date][id_col].unique()
        else:
            append_ids = p_history[id_col].unique()
        if exclude_ids is not None:  # exclude specified ids
            append_ids = np.array([x for x in append_ids if x not in exclude_ids])
        if include_ids is not None:  # append ids that should be included
            append_ids = np.array(list(append_ids) + include_ids)
        length = append_ids.size
        return pd.DataFrame(
            {
                datetime_col: [the_date] * length,
                id_col: append_ids,
                "transactions": [0] * length,
            },
        ).set_index(datetime_col)

    if append_row is not None:
        if verbose:
            print("Finding accounts with no transactions in the period...")
        no_sales_ids = [
            x for x in append_row.index if x not in sales["ZCUSTOMER"].unique()
        ]

        if verbose:
            print(
                "Finding [new] accounts with their first transaction in the period..."
            )
        new_ids = [
            x for x in sales["ZCUSTOMER"].unique() if x not in append_row.index.unique()
        ]

    else:
        no_sales_ids = None
        new_ids = None

    p_history = p_history.append(get_date(end_date, include_ids=no_sales_ids))
    p_history = p_history.append(
        get_date(start_date, include_ids=no_sales_ids, exclude_ids=new_ids)
    )

    p_history = p_history.groupby(id_col).resample("D").sum()  # slow

    p_history["T"] = p_history.groupby(id_col).cumcount()

    p_history["transactions"] = p_history["transactions"].astype(bool)

    if ignore_first_purchase:
        adj = 1
    else:
        adj = 0
    p_history["frequency"] = p_history.groupby(id_col)["transactions"].cumsum() - adj

    if append_row is not None:
        tfreq = ["T", "frequency"]
        p_history["T"] = p_history["T"] + 1
        p_history = p_history.join(append_row[tfreq], rsuffix="_first")
        for c in tfreq:
            p_history[c + "_first"] = p_history[c + "_first"].fillna(0)
        for c in tfreq:
            p_history[c] = p_history[c] + p_history[c + "_first"]
        p_history = p_history.drop(columns=[c + "_first" for c in tfreq])

    p_history["recency"] = p_history[p_history["transactions"]]["T"]

    if append_row is not None:
        append_row["Date"] = start_date - timedelta(days=1)
        append_row["transactions"] = None
        append_row = append_row.reset_index().set_index(["ZCUSTOMER", "Date"])
        append_row = append_row[["transactions", "T", "frequency", "recency"]]
        p_history = pd.concat([append_row, p_history]).sort_index()
    p_history["recency"] = (
        p_history.groupby(id_col)["recency"].fillna(method="ffill").fillna(0)
    )

    if append_row is not None:
        p_history = p_history[
            p_history.index.get_level_values("Date") != start_date - timedelta(days=1)
        ]

    if tail is not None:
        if isinstance(tail, int):
            tail = end_date - timedelta(days=tail)
        tmp = p_history[p_history.index.get_level_values(datetime_col) >= tail]
    else:
        tmp = p_history

    p_history[p_col] = tmp.apply(
        lambda row: model.conditional_probability_alive(
            row["frequency"], row["recency"], row["T"]
        ),
        axis=1,
    ).apply(lambda x: x[0])

    p_history[p_col] = p_history[p_col].fillna(0)

    p_history["ModelVersion"] = model_version

    for c in ["T", "frequency", "recency"]:
        p_history[c] = p_history[c].astype(int)

    if filename is not None:
        _save_dataframe(p_history, filename)

    if verbose:
        print(f"  Elapsed: {datetime.now() - timer}")

    p_history["SalesOrg_Name"] = SALES_ORG_NAME

    return p_history

def main():
    print('You are here @ scoring')

if __name__ == '__main__':
    main()