import re
import os
import pyodbc
import urllib
import pandas as pd
from datetime import date
from datetime import datetime
from typing import Union
from lifetimes import BetaGeoFitter
from lifetimes import GammaGammaFitter
from sqlalchemy import create_engine
from sqlalchemy.sql import text

SALES_ORG_NAME = os.environ.get("SALES_ORG_NAME")
DATABASE = os.environ.get("ACE_DEV_SQL_DATABASE_LANDING")
DB_PASSWORD = os.environ.get("ACE_DEV_SQL_PASSWORD_LANDING")
DB_USERID = os.environ.get("ACE_DEV_SQL_USERNAME_LANDING")
MODEL_VERSION = os.environ.get("MODEL_VERSION")
RFM_DAYS = int(os.environ.get("RFM_DAYS"))
SERVER = os.environ.get("ACE_DEV_SQL_SERVER")
DRIVER = '{ODBC Driver 17 for SQL Server}'

def get_engine():
    params = urllib.parse.quote_plus(f'Driver=ODBC Driver 17 for SQL Server;Server=tcp:{SERVER},1433;Database={DATABASE};Uid={DB_USERID};Pwd={DB_PASSWORD};Encrypt=yes;TrustServerCertificate=no;Connection Timeout=30;')
    conn_str = 'mssql+pyodbc:///?odbc_connect={}'.format(params)
    return create_engine(conn_str, echo=False, fast_executemany=True)

def get_max_sales_date(
    bw: bool = False, to_datetime: bool = False, verbose: bool = True
) -> Union[date, datetime]:
    """
    Get the date of the most recent sales transaction.

    Sometimes there are transactions with future dates, so don't let this date
    go beyond yesterday. Also remove Invoice Contract (FV) and Debit Memo (L2),
    as these often occur in the future.

    Args:
        bw: Query BW instead of Sherlock.
        to_datetime: If the result is a datetime.date, convert it to a
            datetime.
    """
    if verbose:
        print("Finding max sales date... ")

    max_sales = read(
        f"""
        SELECT MAX([Date]) AS [Date]
        FROM view_CHM_CopaSales
        WHERE SalesOrg_Name = '{SALES_ORG_NAME}'
            AND BILL_TYPE NOT IN ('FV', 'L2')
        """,
        verbose=False,
    )["Date"][0]

    if to_datetime and type(max_sales) == date:
        max_sales = datetime.combine(max_sales, datetime.min.time())

    if verbose:
        print(f"  -> {max_sales}")
    return max_sales

def get_max(
    col_name: str = "[Date]",
    table_name: str = "view_CHM_P_min",
    func: str = "max",
    to_datetime: bool = False,
    time: bool = False,  # passed to sql.read
    verbose: bool = True,  # passed to sql.read
    **kwargs,  # passed to sql.read -> pd.read_sql
) -> Union[datetime, date]:
    """This gets the max value from a table and column.
    Args:
        col_name:
        table_name:
        func: The SQL command to use. For example, this could be used to get
            the min or count instead.
        to_datetime: If the result is a datetime.date, convert it to a
            datetime.

    """
    if table_name.startswith("BW"):
        salesorg = sql_get_salesorg()
        salesorg_col = "SALESORG"
    else:
        salesorg = SALES_ORG_NAME
        salesorg_col = "SalesOrg_Name"
    message = f"SELECT {func.upper()}({col_name}) FROM {table_name}... "
    val = read(
        f"""
        SELECT {func}({col_name})
        FROM {table_name}
        WHERE {salesorg_col} = '{salesorg}'
        """,
        show_query=False,
        message=message,
        message_end="" if verbose else None,
        time=time,
        verbose=verbose,
        **kwargs,
    ).iloc[0][0]

    if to_datetime and type(val) == date:
        val = datetime.combine(val, datetime.min.time())

    if verbose:
        print(f"{val}")
        if val is None:
            print(f"Warning: {message} is None!")
    return val

def get_salesorg() -> str:
    return read(
        f"""
        SELECT SALESORG
        FROM CHM_SalesOrg_to_Name
        WHERE SalesOrg_Name = '{SALES_ORG_NAME}'
        """,
        message="Getting SALESORG...",
        show_query=False,
        time=False,
    ).loc[0][0]

def read(
    query: str,
    top: str = None,
    message: str = None,
    message_end: str = None,
    time: bool = True,
    show_query: bool = True,
    verbose: bool = True,
    **kwargs,  # passed to pd.read_sql
):
    if top is not None:
        query = re.sub(r"^\s*SELECT", f"SELECT TOP {top} ", query)
    if message is None:
        message = f"Downloading {SALES_ORG_NAME} data from {SERVER}... "
        message_end = None
    if verbose:
        print(message, end=message_end)
        if show_query:
            print(query)
        if time:
            t_start = datetime.now()
            print(f"  Started: {t_start}")
    df = pd.read_sql(query, get_engine(), **kwargs)
    if verbose & time:
        print(f"    Elapsed: {datetime.now() - t_start}")
    return df

def get_model_alive(model_version: str = MODEL_VERSION):

    params_ = pd.read_sql(
        f"""
        SELECT r, alpha, a, b
        FROM CHM_Models_BGNBD_Alive
        WHERE Version = '{model_version}'
        """,
        get_engine(),
    ).iloc[0]
    params_.name = None
    model_alive = BetaGeoFitter()
    model_alive.params_ = params_
    model_alive.predict = (
        model_alive.conditional_expected_number_of_purchases_up_to_time
    )
    return model_alive

def append(
    table_name: str,
    df: Union[pd.DataFrame, pd.Series],
    map_cols: bool = False,
    verbose: bool = True,
    **kwargs,
):
    t_start = datetime.now()
    if verbose:
        print(f"> Appending to the {table_name} table...")
        print(f"  Started: {t_start}")
    df = df.reset_index()
    if map_cols:
        df = column_mapper(df)
    df.to_sql(
        table_name, get_engine(), index=False, if_exists="append", **kwargs,
    )
    if verbose:
        print(f"  Elapsed: {datetime.now() - t_start}")
    return True

def execute(query: str, show_query: bool = False, verbose: bool = True):
    """Run an arbitrary SQL query."""
    if verbose:
        print("Executing query... ", end="")
        if show_query:
            print(query)
    _ = con.execute(query)
    if verbose:
        print("Done")

def get_territory_filter(prefix: str = "", col_name: str = "TerritoryName__c"):
    pattern = None
    if SALES_ORG_NAME == "PHUS":
        pattern = "[A][A0-8][0-9]%"
    elif SALES_ORG_NAME == "PHDE":
        pattern = "[G][E0-8][0-9]%"
    elif SALES_ORG_NAME == "PHCA":
        pattern = "[W][0-7][0-9]%"
    if pattern is not None:
        return f" AND {prefix}{col_name} COLLATE Latin1_General_BIN LIKE '{pattern}'"
    return ""

def get_model_purchase(model_version: str = MODEL_VERSION):
    params_ = pd.read_sql(
        f"""
        SELECT r, alpha, a, b
        FROM CHM_Models_BGNBD_Purchase
        WHERE Version = '{model_version}'
        """,
        get_engine(),
    ).iloc[0]
    params_.name = None
    model_purchase = BetaGeoFitter()
    model_purchase.params_ = params_
    model_purchase.predict = (
        model_purchase.conditional_expected_number_of_purchases_up_to_time
    )
    return model_purchase

def main():
    print('You are here @ sql')

if __name__ == '__main__':
    main()